# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""HTTP enabled process executor."""

import socket
import sys

if sys.version_info.major == 2:
    from httplib import HTTPConnection, HTTPException, OK
    from urlparse import urlparse
else:
    from http.client import HTTPConnection, HTTPException, OK
    from urllib.parse import urlparse

from mirakuru.tcp import TCPExecutor


class HTTPExecutor(TCPExecutor):

    """Http enabled process executor."""

    def __init__(self, command, url, shell=False, timeout=None, sleep=0.1):
        """
        Initialize HTTPExecutor executor.

        :param str command: command to run to start service
        :param str url: url where executor can check
            if process has already started.
        :param bool shell: see `subprocess.Popen`
        :param int timeout: time to wait for process to start or stop.
            if None, wait indefinitely.
        :param float sleep: how often to check for start/stop condition
        """
        self.url = urlparse(url)
        """
        An :func:`urlparse.urlparse` representation of an url.

        It'll be used to check process status on."""
        TCPExecutor.__init__(
            self, command, host=self.url.hostname,
            port=self.url.port, shell=shell, timeout=timeout, sleep=sleep)

    def start(self):
        """Start process and wait for sucessful head on defined url."""
        TCPExecutor.start(self)
        self.wait_for(self._wait_for_successful_head)

    def _wait_for_successful_head(self):
        """Check if defined url returns successful head."""
        try:
            conn = HTTPConnection(self.url.hostname,
                                  self.url.port)

            conn.request('HEAD', self.url.path)
            response = conn.getresponse()

            if response.status is OK:
                conn.close()
                return True

        except (HTTPException, socket.timeout):
            return False



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""This executor awaits for appearance of a predefined banner in output."""

import re
from mirakuru.base import Executor


class OutputExecutor(Executor):

    """Executor that awaits for string output being present in output."""

    def __init__(self, command, banner, shell=False, timeout=None, sleep=0.1):
        """
        Initialize OutputExecutor executor.

        :param str command: command to run to start service
        :param str banner: string that has to appear in process output -
            should compile to regular expression.
        :param bool shell: see `subprocess.Popen`
        :param int timeout: time to wait for process to start or stop.
            if None, wait indefinitely.
        :param float sleep: how often to check for start/stop condition
        """
        Executor.__init__(self, command, shell, timeout, sleep)
        self._banner = re.compile(banner)

    def start(self):
        """
        Start process.

        .. note::

            Process will be considered started, when defined banner will appear
            in process output.
        """
        Executor.start(self)
        self.wait_for(self._wait_for_output)

    def _wait_for_output(self):
        """Check if output matches banner."""
        if self._banner.match(self.output().readline()):
            return True
        return False



"""Mirakuru's exceptions."""


class TimeoutExpired(Exception):

    """Is raised when the timeout expires while starting an executor."""

    def __init__(self, executor, timeout):
        """
        Exception initialization.

        :param mirakuru.base.Executor executor: executor for
            which exception occured
        :param int timeout: timeout for which exception occurred.
        """
        self.executor = executor
        self.timeout = timeout

    def __str__(self):
        """
        Return Exception's string representation.

        :returns: string representation
        :rtype: str
        """
        return 'Executor {0} timed out after {1} seconds'.format(
            self.executor, self.timeout
        )



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""Base executor with the most basic functionality."""

import subprocess
import shlex
import time

from mirakuru.exceptions import TimeoutExpired


class Executor(object):

    """Basic executor with the most basic functionality."""

    def __init__(self, command, shell=False, timeout=None, sleep=0.1):
        """
        Initialize executor.

        :param str command: command to run to start service
        :param bool shell: see `subprocess.Popen`
        :param int timeout: time to wait for process to start or stop.
            if None, wait indefinitely.
        :param float sleep: how often to check for start/stop condition

        .. note::

            **timeout** set for executor is valid for all the level of waits
            on the way up. That means that if some more advanced executor sets
            timout to 10 seconds, and it'll take 5 seconds for first check,
            second check will only have 5 seconds left.

        """
        self._args = shlex.split(command)
        self._shell = shell
        self._timeout = timeout
        self._sleep = sleep

        self._endtime = None
        self.process = None
        """A :class:`subprocess.Popen` instance once process is started."""

    def running(self):
        """
        Check if executor is running.

        :returns: True if process is running, False otherwise
        :rtype: bool
        """
        if self.process is None:
            return False
        else:
            return self.process.poll() is None

    def start(self):
        """
        Start defined process.

        After process gets started, timeout countdown begins as well.

        .. note::
            We want to open ``stdin``, ``stdout`` and ``stderr`` as text
            streams in universal newlines mode, so we have to set
            ``universal_newlines`` to ``True``.
        """
        if self.process is None:
            self.process = subprocess.Popen(
                self._args,
                shell=self._shell,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                universal_newlines=True,
            )

        self._set_timeout()

    def _set_timeout(self, timeout=None):
        """
        Set timout for possible wait.

        :param int timeout: [optional] specific timeout to set.
            If not set, Executor._timeout will be used instead.
        """
        timeout = timeout or self._timeout

        if timeout:
            self._endtime = time.time() + timeout

    def stop(self):
        """
        Stop process running.

        Wait 10 seconds for the process to end, then just kill it.

        .. note::

            When gathering coverage for the subprocess in tests,
            you have to allow subprocesses to end gracefully.
        """
        if self.process is not None:
            self.process.terminate()

            def process_stopped():
                return self.running() is False

            # set 10 seconds wait no matter what to kill the process
            self._set_timeout(10)
            try:
                self.wait_for(process_stopped)
            except TimeoutExpired:
                # at this moment, process got killed,
                pass

            self.process = None
            self._endtime = None

    def kill(self, wait=True):
        """
        Kill the process if running.

        :param bool wait: set to `True` to wait for the process to end,
            or False, to simply proceed after sending signal.
        """
        if self.running():
            self.process.kill()
            if wait:
                self.process.wait()
            self.process = None
            self._endtime = None

    def output(self):
        """Return process output."""
        if self.process is not None:
            return self.process.stdout

    def wait_for(self, wait_for):
        """
        Wait for callback to return True.

        Simply returns if wait_for condition has been met,
        raises TimeoutExpired otherwise and kills the process.

        :param callback wait_for: callback to call
        :raises: mirakuru.exceptions.TimeoutExpired
        """
        while self.check_timeout():
            if wait_for():
                return
            time.sleep(self._sleep)

        self.kill()
        raise TimeoutExpired(
            self, timeout=self._timeout
        )

    def check_timeout(self):
        """
        Check if timeout has expired.

        Returns True if there is no timeout set or the timeout has not expired.
        Kills the process and raises TimeoutExpired exception otherwise.

        This method should be used in while loops waiting for some data.

        :returns: True if timeout expired, False if not
        :rtype: bool
        """
        if self._endtime is not None and time.time() > self._endtime:
            return False
        return True
